/*
 * This file is part of unity-2d
 *
 * Copyright 2010 Canonical Ltd.
 *
 * Authors:
 * - Aurélien Gâteau <aurelien.gateau@canonical.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Self
#include "unity2dpanel.h"
#include <debug_p.h>

// Qt
#include <QApplication>
#include <QDesktopWidget>
#include <QPainter>
#include <QPropertyAnimation>
#include <QHBoxLayout>
#include <QX11Info>
#include "gconfitem-qml-wrapper.h"

// X
#include <X11/Xlib.h>
#include <X11/Xatom.h>

static const int SLIDE_DURATION = 500;

struct Unity2dPanelPrivate
{
    Unity2dPanel* q;
    Unity2dPanel::Edge m_edge;
    QHBoxLayout* m_layout;
    QPropertyAnimation* m_slideInAnimation;
    QPropertyAnimation* m_slideOutAnimation;
    bool m_useStrut;
    int m_delta;
    bool m_manualSliding;
    bool m_showBackground;
    GConfItemQmlWrapper m_backgroundFilename;
    GConfItemQmlWrapper m_backgroundOptions;
    QImage m_backgroundImage;

    void setStrut(ulong* struts)
    {
        static Atom atom = XInternAtom(QX11Info::display(), "_NET_WM_STRUT_PARTIAL", False);
        XChangeProperty(QX11Info::display(), q->effectiveWinId(), atom,
                        XA_CARDINAL, 32, PropModeReplace,
                        (unsigned char *) struts, 12);
    }

    void reserveStrut()
    {
        QDesktopWidget* desktop = QApplication::desktop();
        const QRect screen = desktop->screenGeometry(q);
        const QRect available = desktop->availableGeometry(q);

        ulong struts[12];
        switch (m_edge) {
        case Unity2dPanel::LeftEdge:
            struts = {
                q->width(), 0, 0, 0,
                available.top(), available.bottom(), 0, 0,
                0, 0, 0, 0
                };
            break;
        case Unity2dPanel::TopEdge:
            struts = {
                0, 0, q->height(), 0,
                0, 0, 0, 0,
                screen.left(), screen.right(), 0, 0
                };
            break;
        }

        setStrut(struts);
    }

    void releaseStrut()
    {
        ulong struts[12];
        memset(struts, 0, sizeof struts);
        setStrut(struts);
    }

    void updateGeometry()
    {
        QDesktopWidget* desktop = QApplication::desktop();
        const QRect screen = desktop->screenGeometry(q);
        const QRect available = desktop->availableGeometry(q);

        QRect rect;
        switch (m_edge) {
        case Unity2dPanel::LeftEdge:
            rect = QRect(screen.left(), available.top(), q->width(), available.height());
            rect.moveLeft(m_delta);
            break;
        case Unity2dPanel::TopEdge:
            rect = QRect(screen.left(), screen.top(), screen.width(), q->height());
            rect.moveTop(m_delta);
            break;
        }

        q->setGeometry(rect);
        loadBackground();
        q->update();
    }

    void updateLayoutDirection()
    {
        QBoxLayout::Direction direction;
        switch(m_edge) {
        case Unity2dPanel::TopEdge:
            direction = QApplication::isRightToLeft() ? QBoxLayout::RightToLeft : QBoxLayout::LeftToRight;
            break;
        case Unity2dPanel::LeftEdge:
            direction = QBoxLayout::TopToBottom;
            break;
        }
        m_layout->setDirection(direction);
    }

    void updateEdge()
    {
        if (m_useStrut) {
            reserveStrut();
        }
        updateGeometry();
        updateLayoutDirection();
    }

    void loadBackground()
    {
        QString options = m_backgroundOptions.getValue().toString();
        QString fileName = m_backgroundFilename.getValue().toString();
        if (!m_backgroundImage.load(fileName)) {
            // FIXME: need more proper error handling during drawing if this failed
            qWarning() << "Failed to load background image:" << fileName;
        } else {
            QRect screen = QApplication::desktop()->screenGeometry(QX11Info::appScreen());
            if (options == "zoom") {
                // FIXME: this won't work for backgrounds that are not horizontal or screens that are not horizontal
                m_backgroundImage = m_backgroundImage.scaledToWidth(screen.width(), Qt::SmoothTransformation);
                m_backgroundImage = m_backgroundImage.copy(0, (m_backgroundImage.height() - screen.height()) / 2,
                                                           screen.width(), q->height());
            } else /* if (options == "scaled") */ {
                m_backgroundImage = m_backgroundImage.scaled(QSize(screen.width(), screen.height()),
                                                             Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
                m_backgroundImage = m_backgroundImage.copy(0, 0, screen.width(), q->height());
            }
        }
        Q_EMIT q->backgroundChanged(fileName, options);
    }
};

Unity2dPanel::Unity2dPanel(QWidget* parent)
: QWidget(parent)
, d(new Unity2dPanelPrivate)
{
    d->q = this;
    d->m_edge = Unity2dPanel::TopEdge;
    d->m_useStrut = true;
    d->m_delta = 0;
    d->m_manualSliding = false;
    d->m_layout = new QHBoxLayout(this);
    d->m_layout->setMargin(0);
    d->m_layout->setSpacing(0);

    d->m_slideInAnimation = new QPropertyAnimation(this);
    d->m_slideInAnimation->setTargetObject(this);
    d->m_slideInAnimation->setPropertyName("delta");
    d->m_slideInAnimation->setDuration(SLIDE_DURATION);
    d->m_slideInAnimation->setEasingCurve(QEasingCurve::InOutCubic);
    d->m_slideInAnimation->setEndValue(0);

    d->m_slideOutAnimation = new QPropertyAnimation(this);
    d->m_slideOutAnimation->setTargetObject(this);
    d->m_slideOutAnimation->setPropertyName("delta");
    d->m_slideOutAnimation->setDuration(SLIDE_DURATION);
    d->m_slideOutAnimation->setEasingCurve(QEasingCurve::InOutCubic);
    d->m_slideOutAnimation->setEndValue(-panelSize());

    d->m_showBackground = false;

    setAttribute(Qt::WA_X11NetWmWindowTypeDock);
    setAttribute(Qt::WA_Hover);
    setAutoFillBackground(true);
    connect(QApplication::desktop(), SIGNAL(workAreaResized(int)), SLOT(slotWorkAreaResized(int)));
}

Unity2dPanel::~Unity2dPanel()
{
    if (d->m_useStrut) {
        d->releaseStrut();
    }
    delete d;
}

void Unity2dPanel::setEdge(Unity2dPanel::Edge edge)
{
    d->m_edge = edge;
    if (isVisible()) {
        d->updateEdge();
    }
}

Unity2dPanel::Edge Unity2dPanel::edge() const
{
    return d->m_edge;
}

bool Unity2dPanel::showBackground() const
{
    return d->m_showBackground;
}

void Unity2dPanel::setShowBackground(bool showBackground)
{
    if (showBackground != d->m_showBackground) {
        if (showBackground) {
            d->m_backgroundFilename.setKey("/desktop/gnome/background/picture_filename");
            d->m_backgroundOptions.setKey("/desktop/gnome/background/picture_options");
            connect(&d->m_backgroundFilename, SIGNAL(valueChanged()), SLOT(updateBackground()));
            connect(&d->m_backgroundOptions, SIGNAL(valueChanged()), SLOT(updateBackground()));
        } else {
            d->m_backgroundFilename.disconnect(this);
            d->m_backgroundOptions.disconnect(this);
        }
        d->m_showBackground = showBackground;
        Q_EMIT showBackgroundChanged(showBackground);

        updateBackground();
    }
}

void Unity2dPanel::updateBackground()
{
    d->loadBackground();
    update();
}

void Unity2dPanel::showEvent(QShowEvent* event)
{
    QWidget::showEvent(event);
    d->updateEdge();
    d->m_slideOutAnimation->setEndValue(-panelSize());
}

void Unity2dPanel::resizeEvent(QResizeEvent* event)
{
    QWidget::resizeEvent(event);
    d->m_slideOutAnimation->setEndValue(-panelSize());
}

void Unity2dPanel::slotWorkAreaResized(int screen)
{
    if (x11Info().screen() == screen) {
        d->updateEdge();
    }
}

void Unity2dPanel::paintEvent(QPaintEvent* event)
{
    // Necessary because Oxygen thinks it knows better what to paint in the background
    QPainter painter(this);
    painter.setCompositionMode(QPainter::CompositionMode_Source);
    painter.fillRect(rect(), palette().brush(QPalette::Background));

    if (d->m_showBackground && !d->m_backgroundImage.isNull()) {
        QString options = d->m_backgroundOptions.getValue().toString();

        QRect screen = QApplication::desktop()->screenGeometry(QX11Info::appScreen());
        painter.drawImage(QPoint((screen.width() - d->m_backgroundImage.width()) / 2, 0),
                          d->m_backgroundImage);

        QColor overlay = Qt::black;
        overlay.setAlphaF(0.71);
        painter.setCompositionMode(QPainter::CompositionMode_SourceAtop);
        painter.fillRect(rect(), overlay);
    }
}

void Unity2dPanel::addWidget(QWidget* widget)
{
    d->m_layout->addWidget(widget);
}

void Unity2dPanel::addSpacer()
{
    d->m_layout->addStretch();
}

bool Unity2dPanel::useStrut() const
{
    return d->m_useStrut;
}

void Unity2dPanel::setUseStrut(bool value)
{
    if (d->m_useStrut != value) {
        if (value) {
            d->reserveStrut();
        } else {
            d->releaseStrut();
        }
        d->m_useStrut = value;
        Q_EMIT useStrutChanged(value);
    }
}

int Unity2dPanel::delta() const
{
    return d->m_delta;
}

void Unity2dPanel::setDelta(int delta)
{
    /* Clamp delta to be between 0 and minus its size */
    int minDelta = -panelSize();
    int maxDelta = 0;

    d->m_delta = qMax(qMin(delta, maxDelta), minDelta);
    d->updateGeometry();
}

int Unity2dPanel::panelSize() const
{
    return (d->m_edge == Unity2dPanel::TopEdge) ? height() : width();
}

void Unity2dPanel::slideIn()
{
    d->m_slideOutAnimation->stop();
    if (d->m_slideInAnimation->state() != QAbstractAnimation::Running) {
        d->m_slideInAnimation->setStartValue(d->m_delta);
        d->m_slideInAnimation->start();
    }
}

void Unity2dPanel::slideOut()
{
    d->m_slideInAnimation->stop();
    if (d->m_slideOutAnimation->state() != QAbstractAnimation::Running) {
        d->m_slideOutAnimation->setStartValue(d->m_delta);
        d->m_slideOutAnimation->start();
    }
}

bool Unity2dPanel::manualSliding() const
{
    return d->m_manualSliding;
}

void Unity2dPanel::setManualSliding(bool manualSliding)
{
    if (d->m_manualSliding != manualSliding) {
        d->m_manualSliding = manualSliding;
        if (manualSliding) {
            d->m_slideInAnimation->stop();
            d->m_slideOutAnimation->stop();
        }
        Q_EMIT manualSlidingChanged(d->m_manualSliding);
    }
}

#include "unity2dpanel.moc"
