/*
 * This file is part of unity-2d
 *
 * Copyright 2010 Canonical Ltd.
 *
 * Authors:
 * - Aurélien Gâteau <aurelien.gateau@canonical.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Local
#include <config.h>

// Applets
#include <appindicator/appindicatorapplet.h>
#include <appname/appnameapplet.h>
#include <homebutton/homebuttonapplet.h>
#include <indicator/indicatorapplet.h>
#include <legacytray/legacytrayapplet.h>

// Unity
#include <gnomesessionclient.h>
#include <unity2ddebug.h>
#include <unity2dpanel.h>
#include <unity2dapplication.h>
#include <unity2dstyle.h>
#include <unity2dtr.h>

// Qt
#include <QAbstractFileEngineHandler>
#include <QApplication>
#include <QFSFileEngine>
#include <QLabel>
#include <QPainter>

using namespace Unity2d;

class ThemeEngineHandler : public QAbstractFileEngineHandler
{
public:
    QAbstractFileEngine *create(const QString& fileName) const
    {
        if (fileName.startsWith("theme:")) {
            QString name = UNITY_DIR "themes/" + fileName.mid(6);
            return new QFSFileEngine(name);
        } else {
            return 0;
        }
    }
};

QPalette getPalette()
{
    QPalette palette;
    palette.setBrush(QPalette::Window, Qt::transparent);
    palette.setBrush(QPalette::Button, Qt::transparent);
    palette.setColor(QPalette::WindowText, Qt::white);
    palette.setColor(QPalette::ButtonText, Qt::white);
    return palette;
}

static const QSize SEPARATOR_SIZE = QSize(6, 24);

QLabel* createSeparator()
{
    QPixmap separator(SEPARATOR_SIZE);
    separator.fill(Qt::transparent);
    QPainter painter(&separator);

    QPixmap divider(unity2dDirectory() + "/panel/artwork/divider.png");
    painter.drawPixmap(QPoint(0, (separator.height() - divider.height()) / 2), divider);

    QLabel* label = new QLabel;
    label->setPixmap(separator);
    label->setFixedSize(separator.size());
    return label;
}

int main(int argc, char** argv)
{
    ThemeEngineHandler handler;

    Unity2dDebug::installHandlers();

    /* Forcing graphics system to 'raster' instead of the default 'native'
       which on X11 is 'XRender'.
       'XRender' defaults to using a TrueColor visual. We mimick that behaviour
       with 'raster' by calling QApplication::setColorSpec.

       Reference: https://bugs.launchpad.net/upicek/+bug/674484
    */
    QApplication::setGraphicsSystem("raster");
    QApplication::setColorSpec(QApplication::ManyColor);
    Unity2dApplication app(argc, argv);
    QApplication::setStyle(new Unity2dStyle);

    GnomeSessionClient client(INSTALL_PREFIX "/share/applications/unity-2d-panel.desktop");
    client.connectToSessionManager();

    /* Configure translations */
    Unity2dTr::init("unity-2d", INSTALL_PREFIX "/share/locale");

    Unity2dPanel panel;
    panel.setEdge(Unity2dPanel::TopEdge);
    panel.setPalette(getPalette());
    panel.setFixedHeight(24);
    panel.setShowBackground(true);

    panel.addWidget(new HomeButtonApplet);
    panel.addWidget(createSeparator());
    panel.addWidget(new AppNameApplet);

    LegacyTrayApplet* trayApplet = new LegacyTrayApplet();
    trayApplet->connect(&panel, SIGNAL(backgroundChanged(QString,QString)),
                        SLOT(updateTrayIcons()));
    panel.addWidget(trayApplet);

    IndicatorApplet* indicatorApplet = new IndicatorApplet();
    indicatorApplet->connect(&panel, SIGNAL(backgroundChanged(QString,QString)),
                             SLOT(updateBackground(QString,QString)));
    panel.addWidget(indicatorApplet);

    panel.show();
    return app.exec();
}
