/*
 * This file is part of unity-2d
 *
 * Copyright 2010 Canonical Ltd.
 *
 * Authors:
 * - Aurélien Gâteau <aurelien.gateau@canonical.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// Self
#include "indicatorapplet.h"

// Local
#include "abstractindicator.h"
#include "datetimeindicator.h"
#include "debug_p.h"
#include "indicator.h"

// Qt
#include <QAction>
#include <QDBusConnection>
#include <QHBoxLayout>
#include <QMenu>
#include <QX11EmbedContainer>
#include <QX11Info>
#include <QApplication>
#include <QDesktopWidget>
#include <QPoint>

// Gtk
#undef signals
#include <gdk/gdk.h>
#include <gtk/gtk.h>

IndicatorApplet::IndicatorApplet() : m_gtkIndicator(NULL)
{
    setupUi();
    loadIndicators();
}

void IndicatorApplet::setupUi()
{
    m_menuBar = new QMenuBar;
    m_menuBar->setNativeMenuBar(false);
    QHBoxLayout* layout = new QHBoxLayout(this);
    layout->setMargin(0);
    layout->addWidget(m_menuBar);

    QMetaObject::invokeMethod(this, "createGtkIndicator", Qt::QueuedConnection);
}

void IndicatorApplet::createGtkIndicator()
{
    int* argc = 0;
    char*** argv = 0;
    gtk_init(argc, argv);

    m_container = new QX11EmbedContainer;
    layout()->addWidget(m_container);

    m_gtkIndicator = indicator_new();
    m_container->embedClient(gtk_plug_get_id(GTK_PLUG(m_gtkIndicator->container)));
    gtk_widget_show(m_gtkIndicator->container);

    QTimer* timer = new QTimer(this);
    timer->setInterval(1000);
    timer->setSingleShot(false);
    connect(timer, SIGNAL(timeout()), SLOT(adjustGtkIndicatorSize()));
    timer->start();
}

void IndicatorApplet::adjustGtkIndicatorSize()
{
    GtkRequisition requisition;
    gtk_widget_size_request(m_gtkIndicator->menu, &requisition);

    /* FIXME: this should happen only if the size actually changed, otherwise
       we will waste a lot of resources continually setting the background over
       and over */
    m_container->setFixedWidth(requisition.width);
    updateBackgroundFromStrip();
}

void IndicatorApplet::loadIndicators()
{
#if 0
    // FIXME: Using Qt plugins
    QList<AbstractIndicator*> indicators = QList<AbstractIndicator*>()
        << new DateTimeIndicator(this)
        ;

    Q_FOREACH(AbstractIndicator* indicator, indicators) {
        connect(indicator, SIGNAL(actionAdded(QAction*)), SLOT(slotActionAdded(QAction*)));
        connect(indicator, SIGNAL(actionRemoved(QAction*)), SLOT(slotActionRemoved(QAction*)));
        indicator->init();
    }
#endif
}

void IndicatorApplet::slotActionAdded(QAction* action)
{
    UQ_VAR(action->text());
    m_menuBar->addAction(action);
}

void IndicatorApplet::slotActionRemoved(QAction* action)
{
    m_menuBar->removeAction(action);
}

/* Load the background with the requested scaling option,
   then keep one strip as wide as the screen and as tall as the panel, with opacity already
   applied.
   This strip will be used to update the actual background of the indicators widget.
   The point of keeping the entire strip around is to avoid having to reload the background
   whenever the size of the indicators widget changes.
*/
void IndicatorApplet::updateBackground(QString fileName, QString options)
{
    if (m_gtkIndicator == NULL) {
        return;
    }

    /* Clean up any previous background and/or strip */
    if (m_gtkIndicator->background != NULL) {
        g_object_unref(m_gtkIndicator->background);
        m_gtkIndicator->background = NULL;
    }
    if (m_gtkIndicator->backgroundStrip != NULL) {
        g_object_unref(m_gtkIndicator->backgroundStrip);
        m_gtkIndicator->backgroundStrip = NULL;
    }

    QRect screen = QApplication::desktop()->screenGeometry(QX11Info::appScreen());
    GdkScreen *gdkscreen = gdk_screen_get_default();
    int depth = gdk_screen_get_system_visual(gdkscreen)->depth;
    GError *error = NULL;

    /* Allocate a new empty strip of the right size */
    m_gtkIndicator->backgroundStrip = gdk_pixmap_new(NULL, screen.width(), height(),
                                                     depth);
    gdk_drawable_set_colormap(GDK_DRAWABLE(m_gtkIndicator->backgroundStrip),
                              gdk_screen_get_default_colormap(gdkscreen));
    cairo_t* cairo;
    GdkPixbuf *fullBackground;

    if (options == "zoom") {
        // FIXME: this won't work for backgrounds that are not horizontal or screens
        //        that are not horizontal
        fullBackground = gdk_pixbuf_new_from_file_at_size(fileName.toUtf8(),
                                                          screen.width(), -1, &error);
    } else {
        fullBackground = gdk_pixbuf_new_from_file_at_scale(fileName.toUtf8(),
                                                           screen.width(), screen.height(),
                                                           FALSE, &error);
    }

    if (error != NULL) {
        qWarning() << "Failed to load background pixmap for indicators" << error->message;
        g_error_free(error);
        return;
    }

    if (options == "zoom") {
        /* Since we scaled up to fill the screen width, the height is now bigger than the screen.
           Therefore we first copy the correct part of the image that should be visible into the
           0, 0 coordinates (effectively shifting it up).
           To optimize this we do the copy on the image itself, and also we copy only as much as
           will be contained inside the strip */
        int scaledHeight = gdk_pixbuf_get_height(fullBackground);
        gdk_pixbuf_copy_area(fullBackground,
                             0, (scaledHeight - screen.height()) / 2,
                             screen.width(), height(),
                             fullBackground, 0, 0);
    }

    /* Finally render the strip into a pixmap, for quicker manipulation and use */
    cairo = gdk_cairo_create(GDK_DRAWABLE(m_gtkIndicator->backgroundStrip));
    gdk_cairo_set_source_pixbuf(cairo, fullBackground, 0, 0);
    cairo_paint(cairo);
    g_object_unref(fullBackground);

    /* Apply a dark layer on top of the background strip */
    cairo_set_source_rgba (cairo, 0.0, 0.0, 0.0, 0.71);
    cairo_set_operator (cairo, CAIRO_OPERATOR_ATOP);
    cairo_rectangle(cairo, 0, 0, screen.width(), height());
    cairo_fill(cairo);
    cairo_destroy(cairo);

    /* If the widget is already visible, proceed to actually setup the background */
    if (width() > 0 && height() > 0) {
        updateBackgroundFromStrip();
    }
}

/* Take the background strip and copy a slice at the top-right that is as wide as
   the indicators widget. Then set it up as the the widget background pixmap */
void IndicatorApplet::updateBackgroundFromStrip()
{
    if (m_gtkIndicator == NULL || m_gtkIndicator->backgroundStrip == NULL) {
        return;
    }

    QRect screen = QApplication::desktop()->screenGeometry(QX11Info::appScreen());
    GdkScreen *gdkscreen = gdk_screen_get_default();
    QPoint globalPos = mapToGlobal(QPoint(0,0));

    /* Clean up any previous background */
    if (m_gtkIndicator->background != NULL) {
        g_object_unref(m_gtkIndicator->background);
        m_gtkIndicator->background = NULL;
    }

    /* Make a temporary copy of the part of the strip that we cover with the widget */
    GdkPixbuf *tmp;
    tmp = gdk_pixbuf_get_from_drawable(NULL, m_gtkIndicator->backgroundStrip,
                                       gdk_drawable_get_colormap(m_gtkIndicator->backgroundStrip),
                                       globalPos.x(), 0, 0, 0, screen.width() - globalPos.x(), height());

    /* Create a new background pixmap suitable for being set as background */
    m_gtkIndicator->background = gdk_pixmap_new(m_gtkIndicator->backgroundStrip, width(), height(),
                                                gdk_screen_get_system_visual(gdkscreen)->depth);
    gdk_drawable_set_colormap(GDK_DRAWABLE(m_gtkIndicator->background),
                              gdk_screen_get_default_colormap(gdkscreen));

    /* Transfer the temporary copy on the pixmap and destroy the copy */
    cairo_t* cairo = gdk_cairo_create(GDK_DRAWABLE(m_gtkIndicator->background));
    gdk_cairo_set_source_pixbuf(cairo, tmp, 0, 0);
    cairo_paint(cairo);
    cairo_destroy(cairo);
    g_object_unref(tmp);

    if (m_gtkIndicator->menu != NULL) {
        /* Set the pixmap as background for the GDK window of the menubar (which is the topmost widget
           in the indicators widget) */
        GtkStyle *style = gtk_style_new();
        style->bg_pixmap[GTK_STATE_NORMAL] = (GdkPixmap*) g_object_ref(m_gtkIndicator->background);
        gtk_style_set_background(style, m_gtkIndicator->menu->window, GTK_STATE_NORMAL);
        g_object_unref(style);

        /* Ask the widget to redraw itself at the next idle loop */
        gtk_widget_queue_draw(m_gtkIndicator->menu);
    }


}

#include "indicatorapplet.moc"
